@@ -1,5 +1,6 @@
 /*
  * Copyright (C) Andrew Tridgell 2002
+ * Copyright (C) Joel Rosdahl 2011
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -170,12 +171,12 @@ execute(char **argv, const char *path_stdout, const char *path_stderr)
 	cc_log_argv("Executing ", argv);
 
 	pid = fork();
-	if (pid == -1) fatal("Failed to fork");
+	if (pid == -1) fatal("Failed to fork: %s", strerror(errno));
 
 	if (pid == 0) {
 		int fd;
 
-		unlink(path_stdout);
+		tmp_unlink(path_stdout);
 		fd = open(path_stdout, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL|O_BINARY, 0666);
 		if (fd == -1) {
 			exit(1);
@@ -183,7 +184,7 @@ execute(char **argv, const char *path_stdout, const char *path_stderr)
 		dup2(fd, 1);
 		close(fd);
 
-		unlink(path_stderr);
+		tmp_unlink(path_stderr);
 		fd = open(path_stderr, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL|O_BINARY, 0666);
 		if (fd == -1) {
 			exit(1);
@@ -195,7 +196,7 @@ execute(char **argv, const char *path_stdout, const char *path_stderr)
 	}
 
 	if (waitpid(pid, &status, 0) != pid) {
-		fatal("waitpid failed");
+		fatal("waitpid failed: %s", strerror(errno));
 	}
 
 	if (WEXITSTATUS(status) == 0 && WIFSIGNALED(status)) {
