@@ -273,7 +273,7 @@ error:
 
 #define WRITE_INT(size, var) \
 	do { \
-		char ch_; \
+		uint8_t ch_; \
 		size_t i_; \
 		for (i_ = 0; i_ < (size); i_++) { \
 			ch_ = ((var) >> (8 * ((size) - i_ - 1))); \
@@ -581,8 +581,6 @@ manifest_put(const char *manifest_path, struct file_hash *object_hash,
 	int ret = 0;
 	int fd1;
 	int fd2;
-	struct stat st;
-	gzFile f1 = NULL;
 	gzFile f2 = NULL;
 	struct manifest *mf = NULL;
 	char *tmp_file = NULL;
@@ -593,38 +591,24 @@ manifest_put(const char *manifest_path, struct file_hash *object_hash,
 	 * not a big deal, and it's also very unlikely.
 	 */
 
-	fd1 = safe_open(manifest_path);
+	fd1 = open(manifest_path, O_RDONLY | O_BINARY);
 	if (fd1 == -1) {
-		cc_log("Failed to open manifest file");
-		goto out;
-	}
-	if (fstat(fd1, &st) != 0) {
-		cc_log("Failed to stat manifest file");
-		close(fd1);
-		goto out;
-	}
-	if (st.st_size == 0) {
 		/* New file. */
 		mf = create_empty_manifest();
 	} else {
-		f1 = gzdopen(fd1, "rb");
+		gzFile f1 = gzdopen(fd1, "rb");
 		if (!f1) {
 			cc_log("Failed to gzdopen manifest file");
 			close(fd1);
 			goto out;
 		}
 		mf = read_manifest(f1);
-		if (!mf) {
-			cc_log("Failed to read manifest file");
-			gzclose(f1);
-			goto out;
-		}
-	}
-
-	if (f1) {
 		gzclose(f1);
-	} else {
-		close(fd1);
+		if (!mf) {
+			cc_log("Failed to read manifest file; deleting it");
+			x_unlink(manifest_path);
+			mf = create_empty_manifest();
+		}
 	}
 
 	if (mf->n_objects > MAX_MANIFEST_ENTRIES) {
